from __future__ import annotations

import sys
from typing import TYPE_CHECKING, NamedTuple

import pytest

from virtualenv.run import session_via_cli
from virtualenv.seed.embed.pip_invoke import PipInvoke
from virtualenv.seed.embed.via_app_data.via_app_data import FromAppData
from virtualenv.seed.seeder import Seeder
from virtualenv.seed.wheels.embed import MIN, OLDEST_SUPPORTED

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture

    from virtualenv.seed.embed.base_embed import BaseEmbed


@pytest.mark.parametrize(
    ("args", "download"),
    [([], False), (["--no-download"], False), (["--never-download"], False), (["--download"], True)],
)
def test_download_cli_flag(args, download, tmp_path) -> None:
    session = session_via_cli([*args, str(tmp_path)])
    assert session.seeder.download is download


@pytest.mark.parametrize("flag", ["--no-wheel", "--wheel=none", "--wheel=embed", "--wheel=bundle"])
def test_wheel_cli_flags_do_nothing(tmp_path, flag) -> None:
    session = session_via_cli([flag, str(tmp_path)])
    if sys.version_info[:2] >= (3, 12):
        expected = {"pip": "bundle"}
    else:
        expected = {"pip": "bundle", "setuptools": "bundle"}
    assert session.seeder.distribution_to_versions() == expected


@pytest.mark.parametrize("flag", ["--no-wheel", "--wheel=none", "--wheel=embed", "--wheel=bundle"])
def test_wheel_cli_flags_warn(tmp_path, flag, capsys) -> None:
    session_via_cli([flag, str(tmp_path)])
    out, err = capsys.readouterr()
    assert "the --wheel and --no-wheel options do nothing" in out + err


def test_unused_wheel_cli_flags_dont_warn(tmp_path, capsys) -> None:
    session_via_cli([str(tmp_path)])
    out, err = capsys.readouterr()
    assert "the --wheel and --no-wheel options do nothing" not in out + err


def test_embed_wheel_versions(tmp_path: Path) -> None:
    session = session_via_cli([str(tmp_path)])
    if sys.version_info[:2] >= (3, 12):
        expected = {"pip": "bundle"}
    else:
        expected = {"pip": "bundle", "setuptools": "bundle"}
    assert session.seeder.distribution_to_versions() == expected


BUNDLED_SEEDERS = [pytest.param(FromAppData, id="app-data"), pytest.param(PipInvoke, id="pip")]


@pytest.mark.parametrize("seeder", BUNDLED_SEEDERS)
def test_bundled_seeder_reports_reason_below_floor(
    seeder: type[BaseEmbed], at_version: Callable[[int, int], MagicMock]
) -> None:
    reason = seeder.cannot_seed(at_version(3, 8))
    assert reason is not None
    assert "Python 3.8" in reason
    assert f"Python {MIN}" in reason
    assert "--no-seed" in reason


@pytest.mark.parametrize("seeder", BUNDLED_SEEDERS)
def test_bundled_seeder_has_no_reason_on_oldest_supported(
    seeder: type[BaseEmbed], at_version: Callable[[int, int], MagicMock]
) -> None:
    assert seeder.cannot_seed(at_version(*OLDEST_SUPPORTED)) is None


def test_base_seeder_never_blocks(at_version: Callable[[int, int], MagicMock]) -> None:
    assert Seeder.cannot_seed(at_version(3, 8)) is None


def test_selection_surfaces_the_seeder_reason(tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch.object(FromAppData, "cannot_seed", return_value="seeder said no for a specific reason")
    with pytest.raises(RuntimeError, match="seeder said no for a specific reason"):
        session_via_cli([str(tmp_path)])


def test_no_seed_bypasses_capability_check(tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch.object(FromAppData, "cannot_seed", return_value="seeder said no")
    assert session_via_cli(["--no-seed", str(tmp_path)]).seeder.enabled is False


@pytest.fixture
def at_version(mocker: MockerFixture) -> Callable[[int, int], MagicMock]:
    def build(major: int, minor: int) -> MagicMock:
        interpreter = mocker.MagicMock()
        interpreter.version_info = _VersionInfo(major, minor)
        return interpreter

    return build


class _VersionInfo(NamedTuple):
    major: int
    minor: int



from __future__ import annotations

import shlex
import shutil
import subprocess
import sys
from argparse import Namespace

import pytest

from virtualenv.activation import BashActivator
from virtualenv.info import IS_WIN
from virtualenv.run import cli_run


@pytest.mark.skipif(IS_WIN, reason="Github Actions ships with WSL bash")
@pytest.mark.parametrize(
    ("tcl_lib", "tk_lib", "present"),
    [
        ("/path/to/tcl", "/path/to/tk", True),
        (None, None, False),
    ],
)
def test_bash_tkinter_generation(tmp_path, tcl_lib, tk_lib, present) -> None:
    # GIVEN
    class MockInterpreter:
        pass

    interpreter = MockInterpreter()
    interpreter.tcl_lib = tcl_lib
    interpreter.tk_lib = tk_lib

    class MockCreator:
        def __init__(self, dest) -> None:
            self.dest = dest
            self.bin_dir = dest / "bin"
            self.bin_dir.mkdir()
            self.interpreter = interpreter
            self.pyenv_cfg = {}
            self.env_name = "my-env"

    creator = MockCreator(tmp_path)
    options = Namespace(prompt=None)
    activator = BashActivator(options)

    # WHEN
    activator.generate(creator)
    content = (creator.bin_dir / "activate").read_text(encoding="utf-8")

    # THEN
    # The teardown logic is always present in deactivate()
    assert "unset _OLD_VIRTUAL_TCL_LIBRARY" in content
    assert "unset _OLD_VIRTUAL_TK_LIBRARY" in content
    assert "unset _OLD_PKG_CONFIG_PATH" in content

    # PKG_CONFIG_PATH is always set
    assert '_OLD_PKG_CONFIG_PATH="${PKG_CONFIG_PATH:-}"' in content
    assert 'PKG_CONFIG_PATH="${VIRTUAL_ENV}/lib/pkgconfig${PKG_CONFIG_PATH:+:${PKG_CONFIG_PATH}}"' in content
    assert "export PKG_CONFIG_PATH" in content
    assert 'PKG_CONFIG_PATH="$_OLD_PKG_CONFIG_PATH"' in content

    if present:
        assert 'if [ /path/to/tcl != "" ]; then' in content
        assert "TCL_LIBRARY=/path/to/tcl" in content
        assert "export TCL_LIBRARY" in content

        assert 'if [ /path/to/tk != "" ]; then' in content
        assert "TK_LIBRARY=/path/to/tk" in content
        assert "export TK_LIBRARY" in content
    else:
        # When not present, the if condition is false, so the block is not executed
        assert "if [ '' != \"\" ]; then" in content, content
        assert "TCL_LIBRARY=''" in content
        # The export is inside the if, so this is fine
        assert "export TCL_LIBRARY" in content


@pytest.mark.skipif(IS_WIN, reason="Github Actions ships with WSL bash")
def test_bash_activate_relocation_resolves_virtual_env(tmp_path, current_fastest) -> None:
    original = tmp_path / "original"
    cli_run([
        "--without-pip",
        str(original),
        "--creator",
        current_fastest,
        "--no-periodic-update",
        "--activators",
        "bash",
    ])
    relocated = tmp_path / "relocated"
    shutil.move(original, relocated)

    work_dir = tmp_path / "workdir"
    work_dir.mkdir()
    activate_script = relocated / "bin" / "activate"
    result = subprocess.run(
        ["bash", "-c", f'source "{activate_script}" 2>/dev/null && echo "$VIRTUAL_ENV"'],
        capture_output=True,
        text=True,
        cwd=str(work_dir),
    )
    assert result.returncode == 0
    assert result.stdout.strip() == str(relocated)


@pytest.mark.skipif(IS_WIN, reason="Github Actions ships with WSL bash")
def test_bash_activate_does_not_export_ps1(tmp_path, current_fastest) -> None:
    dest = tmp_path / "env"
    cli_run([
        "--without-pip",
        str(dest),
        "--creator",
        current_fastest,
        "--no-periodic-update",
        "--activators",
        "bash",
    ])
    activate_script = dest / "bin" / "activate"
    print_ps1 = f"{shlex.quote(sys.executable)} -c 'import os; print(os.environ.get(\"PS1\"))'"
    result = subprocess.run(
        [
            "bash",
            "-c",
            (f"unset PS1; PS1='base$ '; source \"{activate_script}\" && {print_ps1} && deactivate && {print_ps1}"),
        ],
        capture_output=True,
        encoding="utf-8",
        text=True,
    )
    assert result.returncode == 0, result.stderr
    assert result.stdout.splitlines() == ["None", "None"]


@pytest.mark.skipif(IS_WIN, reason="Github Actions ships with WSL bash")
@pytest.mark.parametrize("hashing_enabled", [True, False])
def test_bash(raise_on_non_source_class, hashing_enabled, activation_tester) -> None:
    class Bash(raise_on_non_source_class):
        def __init__(self, session) -> None:
            super().__init__(
                BashActivator,
                session,
                "bash",
                "activate",
                "sh",
                "You must source this script: $ source ",
            )
            self.deactivate += " || exit 1"
            self._invoke_script.append("-h" if hashing_enabled else "+h")

        def activate_call(self, script):
            return super().activate_call(script) + " || exit 1"

        def print_prompt(self):
            return 'printf "%s\\n" "$PS1"'

    activation_tester(Bash)
